import numpy as np
import numba

from .utils import reducedRepConversionMatrices, distancePointToLine, unravel_3d_index

"""
NOTE: 2024-10-15

This algorithm is an attempt to improve the generic hough transform
but it does not work so far. The idea is to better be able
to distinguish line segments as opposed to just lines.
"""


@numba.njit()
def lineIntersection(a1, b1, a2, b2):
    """
    Compute the intersection of two lines:
        a1 + b1*t and a2 + b2*t
    """
    # Compute the cross product
    crossProd = np.cross(b1, b2)
    if crossProd == 0:
        return np.nan
    # Dot the cross product unit vector into the vector from intercept to intercept
    distance = np.dot(a2 - a1, crossProd) / np.sqrt(np.sum(crossProd**2))
    return distance

@numba.njit()
def distancePointToLine(p, a, b):
    """
    Compute intersection of a point (p) with a line (a + bt).

    Parameters
    ----------
    p : numpy.ndarray[3]
        Point to compute distance to the line

    a : numpy.ndarray[3]
        Point located on the line.

    b : numpy.ndarray[3]
        Direction vector of the line.

    Returns
    -------
    d : float
        Distance from the point to the closest point
        on the line.
    """
    t = np.dot(b, p - a)
    closestPointOnLine = a + t*b
    return np.sqrt(np.sum((p - closestPointOnLine)**2))


@numba.njit()
def integratedHough3D(points, directionVectors, averageNNDistance, latticeSize=64, neighborDistance=0.01, lineJumpDistance=5, refinementSteps=2, minPointsPerLine=5):
    """
    Perform line detection on a 3D point cloud.

    This version of the function is follows the algorithm described in
    the original paper [1] (except for step 5):
        1. Have all points vote in the hough space.
        2. Take the highest voted cell as an initial guess for a new line.
        3. Find all points less than a certain distance to the initial line.
        4. Perform linear regression on the nearby points to refine the line.
        5. Update which points are nearby to the line based on the updated line (NEW).
        6. Have all of the points that are part of this line reverse their vote.
        7. Repeat steps 2-6 until there are not enough points left to form a line.

    I found that adding step 5 gave slightly more reliable detections, so
    I have included it. Note that you could take this further, and perform
    several steps of fitting and updating, but here we just do one.

    Parameters
    ----------
    points : numpy.ndarray[N,3]
        Points representing the point cloud.

    directionVectors : numpy.ndarray[M,3]
        Direction vectors represented the discretized directions for
        the lines. Normally generated with `genIcosahedron()`.

    densityField : numpy.ndarray[Q,P,R]
        Course grained density field for the point cloud

    densityFieldCorner : numpy.ndarray[3]
        The point representing the bottom left (minimum) corner of the
        density field, if not [0,0,0].

    latticeSize : int
        The number of discrete boxes to use for the intercept variables in
        hough space.

    neighborDistance : float
        The distance a point can be from the initial guess of a
        line for it to be included in the linear regression
        used to refine the line parameters. Given as fraction of
        the total system size (diagonal length of the bounding box).

    minPointsPerLine : int
        The minimum number of points that must be part of a line
        detection for that line to be included in the final result.

    Returns
    -------
    lineDirections : numpy.ndarray[K,3]
        The vectors giving the directions of detected lines.

    lineAnchorPoints : numpy.ndarray[K,3]
        The position vectors of the anchor point for each line,
        roughly representing the center of the line (but not
        exactly).

    lineLengths : numpy.ndarray[K]
        The length of each line.
    """
    ##############################################
    #              Preparation steps
    ##############################################
    # Shift the mean of the point cloud to the origin
    pointCloudTranslation = np.zeros(3)
    for i in range(3):
        pointCloudTranslation[i] = np.mean(points[:,i])

    points = np.copy(points) - pointCloudTranslation

    # Compute the discrete direction vectors for our lines
    #directionVectors = genIcosahedron(directionGranularity)
    # TODO: Make this generation function numba compatable

    # Find the extents of the point cloud
    bounds = np.zeros((2, 3))
    for i in range(3):
        bounds[:,i] = [np.min(points[:,i]), np.max(points[:,i])]
    #bounds = np.array([ for i in range(3)]).T

    systemLengthScale = np.sqrt(np.sum((bounds[1] - bounds[0])**2))

    stepSize = systemLengthScale / latticeSize

    # Our hough space will be a three dimensional array, with the first two
    # dimensions representing the reduced coordinates x' and y' for the
    # intersection point of a line with a plane that passes through [0,0,0],
    # and the last dimension for the direction vector of the line. Altogether,
    # a point i,j,k in the hough space fully describes a line in the space
    # of the point cloud.
    houghSpaceSize = (latticeSize, latticeSize, len(directionVectors))

    ##############################################
    #    Initial transformation to hough space
    ##############################################
    # The actual hough space where points will be voting
    houghSpace = np.zeros(houghSpaceSize)

    # Now do the voting
    for k in range(len(directionVectors)):
        b = directionVectors[k]
        # We can compute a vector in advance that we just dot with
        # each point to find the x' and y'
        # See equation 2 in Dalitz et al. 2017
        reducedRepMultX, reducedRepMultY = reducedRepConversionMatrices(b)

        # I'm not 100% clear on why adding the system length scale is
        # necessary here, but otherwise most of our lattice represents
        # values of x' and y' that aren't relevant...
        xPrime = np.dot(reducedRepMultX, points.T) + systemLengthScale/2
        yPrime = np.dot(reducedRepMultY, points.T) + systemLengthScale/2
        #xPrime = np.dot(reducedRepMultX, points.T) + systemLengthScale/2
        #yPrime = np.dot(reducedRepMultY, points.T) + systemLengthScale/2

        # Now convert to an index, so we can find the corresponding
        # place on the hough space lattice.
        xIndex = np.floor(xPrime / stepSize).astype(np.uint8)
        yIndex = np.floor(yPrime / stepSize).astype(np.uint8)

        # Get rid of indices that are not on our lattice
        validIndices = np.where((xIndex < latticeSize) & (yIndex < latticeSize))[0]

        #print(validIndices)
        for j in validIndices:
            houghSpace[xIndex[j], yIndex[j], k] += 1

    ##############################################
    #    Iterative selection of lines
    ##############################################
    # Now that we have calulated our hough space, we need to identify
    # maxima in it to find the best lines

    # We do this by starting with the most voted point in hough space
    # and then "unvote" all of the points that are associated with that
    # line, and then repeat the process iteratively

    # # Naively, we just start by taking the maximum cells
    # maxIndices = np.where(houghSpace >= np.max(houghSpace)*selectionThreshold)
    # # Convert to a nicer format, so we can directly index using
    # # maxIndices[i]
    # maxIndices = [(maxIndices[0][i], maxIndices[1][i], maxIndices[2][i]) for i in range(len(maxIndices[0]))]

    # Numba doesn't play well with empty arrays because
    # it can't infer the type of the data, so we just start
    # with an empty line (all zeros) and we will remove
    # it at the end (see return statement).
    # First index is for line, second is for start/end point,
    # and third is for dimension in 3D space
    linePointArr = np.zeros((1, 2, 3))

    while True:
        # The highest voted space
        maxIndex = np.argmax(houghSpace)
        # Now we have to unravel the index, since it is given as a 1D index
        # The function numpy.unravel_index() does this, but it is not supported
        # by numba
        xIndex, yIndex, bIndex = unravel_3d_index(maxIndex, houghSpaceSize)

        # If we don't have enough points, we finish
        # This probably shouldn't be here, because we could
        # have many points in a line that just didn't vote for the
        # specific lattice spot because of discretization artifacts
        if houghSpace[xIndex,yIndex,bIndex] < minPointsPerLine/2:
            print('Exit because not enough votes in highest voted cell')
            break

        #print(maxIndex)
        print(f'Number of votes: {int(houghSpace[xIndex,yIndex,bIndex])}')

        # plt.imshow(houghSpace[:,:,bIndex])
        # plt.colorbar()
        # plt.title(f'b = {bIndex}')
        # plt.show()

        # Converted from index to reduced coordinates
        xPrime = xIndex*stepSize - systemLengthScale/2
        yPrime = yIndex*stepSize - systemLengthScale/2

        # Calculate the anchor point on the line, which,
        # together with the direction vector, fully identifies
        # the line.
        initialDirectionVector = directionVectors[bIndex]
        reducedRepMultX, reducedRepMultY = reducedRepConversionMatrices(initialDirectionVector)
        # See equation 3 from Dalitz et al. 2017
        anchorPoint = xPrime*reducedRepMultX + yPrime*reducedRepMultY

        # Now, we need to refine the line using linear regression.
        # First, that means identifying which points are included in
        # this line.

        # First, identify points that are within a certain distance of the entire
        # line
        distancesToPoints = np.array([distancePointToLine(p, anchorPoint, initialDirectionVector) for p in points])
        nearbyPointIndices = np.where(distancesToPoints < averageNNDistance*neighborDistance)[0]
        nearbyPoints = points[nearbyPointIndices]

        print(f'Number of nearby points: {len(nearbyPoints)}')

        if len(nearbyPoints) < minPointsPerLine:
            print('Exit because not enough nearby points detected')
            break

        # Now, we want to see how points are distributed throughout the
        # space along the line. This lets us distinguish, for example,
        # two distinct line segments that have the same slope and intercept, but
        # are separated by some distance.
        # Parametrize the points along the line
        tArr = np.dot(initialDirectionVector, (nearbyPoints - anchorPoint).T)

        # Weight by the distance from the line
        tArr = tArr# * distancesToPoints[nearbyPointIndices]

        # Convert to a density map along the line
        pathPoints = 50
        pointDensityAlongLine, binEdges = np.histogram(tArr, pathPoints)

        # Can do this because the direction vector is a unit vector
        pathStepSize = (np.max(tArr) - np.min(tArr))/pathPoints

        # Next we need a threshold by which to binarize this point density
        # to indicate whether we are currently next to a line segment or not
        # We can use the average density of points in the volume to predict
        # the number of points we expect on average in a volume of 1/pathPoints**3.
        averageNumberOfPoints = len(points) / pathPoints**3

        # plt.plot(pointDensityAlongLine)
        # plt.axhline(averageNumberOfPoints)
        # plt.show()

        ##############################################
        #    DEBUG
        ##############################################
        # Uncomment these to plot each line after it is detected.
        # Doesn't work with numba, so you'll have to remove the annotation.

        # flattenedLinePoints = [anchorPoint - 10*initialDirectionVector, anchorPoint + 10*initialDirectionVector]

        # lineSet = o3d.geometry.LineSet()
        # lineSet.points = o3d.utility.Vector3dVector(flattenedLinePoints)
        # lineSet.lines = o3d.utility.Vector2iVector([[i,i+1] for i in range(len(flattenedLinePoints))[::2]])

        # pcd = o3d.geometry.PointCloud()
        # pcd.points = o3d.utility.Vector3dVector(points)
        # pcd.paint_uniform_color([0, 1, 0])

        # pcd2 = o3d.geometry.PointCloud()
        # pcd2.points = o3d.utility.Vector3dVector(nearbyPoints)
        # pcd2.paint_uniform_color([1, 0, 0])

        # o3d.visualization.draw_geometries([lineSet, pcd, pcd2])
        ##############################################
        #    DEBUG
        ##############################################

        ##############################################
        #    Compute line cut of density field
        ##############################################
        # Unlike the basic algorithm, here we try to better distinguish
        # if there are multiple, distinct line segments along this line.
        # eg. consider two line segments that are aligned with each other
        # but on opposite sides of the point cloud; these shouldn't be
        # detected as a single line, since the segment in the middle
        # isn't actually a real line segment.

        pointDensityAlongLine[pointDensityAlongLine >= averageNumberOfPoints] = 1
        pointDensityAlongLine[pointDensityAlongLine < 1] = 0

        # Next, we fill in gaps under the allowed distance
        allowedSkipDistance = averageNNDistance*neighborDistance*lineJumpDistance
        allowedSkipDistance /= pathStepSize
        #print(allowedSkipDistance)

        for i in range(len(pointDensityAlongLine)):
            leftAllowedPoints = [i-k for k in range(int(np.floor(allowedSkipDistance))) if i-k >= 0]
            rightAllowedPoints = [i+k for k in range(int(np.floor(allowedSkipDistance))) if i+k < len(pointDensityAlongLine)]

            # If there is a high density point on both sides of the current point,
            # then we can fill in this current point.
            leftBounded = True in [pointDensityAlongLine[lI] == 1 for lI in leftAllowedPoints]
            rightBounded = True in [pointDensityAlongLine[rI] == 1 for rI in rightAllowedPoints]
            if leftBounded and rightBounded:
                pointDensityAlongLine[i] = 1

        # Now we split the line into regions that
        # are above the density threshold, since there could
        # be multiple lines that otherwise would get detected
        # as a single one.
        lineSegmentRegions = []
        inRegion = False
        for i in range(len(pointDensityAlongLine)):
            if pointDensityAlongLine[i] == 1 and not inRegion:
                lineSegmentRegions.append([i, 0])
                inRegion = True

            if pointDensityAlongLine[i] == 0 and inRegion:
                lineSegmentRegions[-1][1] = i-1
                inRegion = False

        # If our region extends to the end, we have to close it
        if inRegion:
            lineSegmentRegions[-1][1] = len(pointDensityAlongLine)-1


        # plt.plot(pointDensityAlongLine)
        # plt.show()

        # If we have no regions, we can't continue
        if len(lineSegmentRegions) == 0:
            print('Exit because no line segments were detected')
            break

        # fig = plt.figure()
        # ax = fig.add_subplot(projection='3d')

        # for j in range(len(lineSegmentRegions)):
        #     # Find which points are relevant to this line segment
        #     # Middle of line segment based on the line cut
        #     middleSegmentIndex = (lineSegmentRegions[j][0] + lineSegmentRegions[j][1]) / 2
        #     lineSegmentCenter = (anchorPoint + np.min(tArr)*initialDirectionVector) + middleSegmentIndex*pathStepSize*initialDirectionVector
        #     # Add a bit of padding to the radius
        #     lineSegmentHalfLength = (lineSegmentRegions[j][1] - lineSegmentRegions[j][0])*pathStepSize*1.1/2
        #     # Looking in a circle is fine because we have already restricted our points to
        #     # be on a line

        #     sphericalDistanceToPoints = np.sqrt(np.sum((lineSegmentCenter - nearbyPoints)**2, axis=-1))
        #     lineSegmentPointIndices = np.where(sphericalDistanceToPoints < lineSegmentHalfLength)[0]
        #     #lineSegmentPointIndices = np.arange(len(nearbyPoints))
        #     lineSegmentPoints = nearbyPoints[lineSegmentPointIndices]
        #     ax.scatter(*lineSegmentPoints.T, alpha=0.1)
        #     ax.scatter(*lineSegmentCenter[None].T)
        #     print(j, len(lineSegmentPoints))

        # plt.show()

        # Now, we need to break these points up into the segments we identified through
        # the density analysis.
        # DEBUG:
        #lineSegmentRegions = [[0, pathPoints-1]]
        #print(lineSegmentRegions)
        for j in range(len(lineSegmentRegions)):

            # plt.plot(pointDensityAlongLine)
            # plt.axvline(lineSegmentRegions[j][0], c='r')
            # plt.axvline(lineSegmentRegions[j][1], c='r')
            # plt.title(j)
            # plt.show()

            # Find which points are relevant to this line segment
            # Middle of line segment based on the line cut
            middleSegmentIndex = (lineSegmentRegions[j][0] + lineSegmentRegions[j][1]) / 2
            lineSegmentCenter = (anchorPoint + np.min(tArr)*initialDirectionVector) + middleSegmentIndex*pathStepSize*initialDirectionVector
            # Add a bit of padding to the radius
            lineSegmentHalfLength = (lineSegmentRegions[j][1] - lineSegmentRegions[j][0])*pathStepSize*1.1/2

            #print(f'Line segment center: {lineSegmentCenter}')
            #print(f'Line segment half width: {lineSegmentHalfWidth}')

            # Looking in a circle is fine because we have already restricted our points to
            # be on a line
            sphericalDistanceToPoints = np.sqrt(np.sum((lineSegmentCenter - nearbyPoints)**2, axis=-1))
            # plt.hist(sphericalDistanceToPoints, bins=30)
            # plt.show()

            lineSegmentPointIndices = np.where(sphericalDistanceToPoints < lineSegmentHalfLength)[0]
            #lineSegmentPointIndices = np.arange(len(nearbyPoints))
            lineSegmentPoints = nearbyPoints[lineSegmentPointIndices]
            #print(f'Number of points in line segment {j}: {len(lineSegmentPoints)}')

            # fig = plt.figure()
            # ax = fig.add_subplot(projection='3d')

            # ax.scatter(*nearbyPoints.T, alpha=0.5)
            # ax.scatter(*lineSegmentPoints.T, alpha=0.5)
            # ax.plot(*np.array([anchorPoint - b*5, anchorPoint + b*5]).T, '--', c='r')

            # plt.show()
            # Remove the used points from the list
            # Indexing is slightly complicated because we did it in two steps
            # points -> nearbyPoints -> lineSegmentPoints
            # So we have to go backwards one step at a time
            # That is; `nearbyPointIndices` gives the indices of points in
            # `points` that are included in `nearbyPoints`, and `lineSegmentPointIndices`
            # gives the indices of points in `nearbyPoints` that are included
            # in `lineSegmentPoints`

            # keepIndices = np.array([i for i in range(len(points)) if i not in nearbyPointIndices[lineSegmentPointIndices]], dtype=np.int64)
            # #print(f'removed {len(points) - len(keepIndices)} points')
            # points = points[keepIndices]

            ##############################################
            #    Refining line parameters
            ##############################################
            # If we have enough points, we continue on to calculate the
            # properties of the line
            # Otherwise, we just unvote these points and move on.
            if len(lineSegmentPoints) >= minPointsPerLine:

                # Now we perform linear regression on those points to fit a line
                # and increase the accuracy of our detected line

                # We will use SVD to find the first principle component
                # of the local points, which is equivalent to a linear
                # fit.
                # To do this, we need our data to be zero-centered
                # (and we need the center of mass point anyway as our
                # anchor point)
                anchorPoint = np.zeros(3)
                for i in range(3):
                    anchorPoint[i] = np.mean(lineSegmentPoints[:,i])

                # SVD on the mean-centered data.
                # The original paper did PCA instead of SVD, as they said it
                # was much faster, but I think SVD is fine here...
                uu, dd, vv = np.linalg.svd(lineSegmentPoints - anchorPoint)
                # Direction vector is the first principle component
                directionVector = vv[0]

                # Compute the length of the line
                # The best way to do this is to compute the parametric representation
                # of every point in lineSegmentPoints. Then we can take the maximum and
                # minimum parametric values, to find how far the line extends in each direction
                tArr = np.dot(directionVector, (lineSegmentPoints - anchorPoint).T)
                lineStart = anchorPoint + np.max(tArr)*directionVector
                lineEnd = anchorPoint + np.min(tArr)*directionVector

                linePoints = np.zeros((2, 3))
                linePoints[0] = lineStart
                linePoints[1] = lineEnd

                # Save the information about this line
                linePointArr = np.concatenate((linePointArr, linePoints[None,:,:]))
                #print(f'Saved information about line segment {j}')


            keepIndices = np.array([i for i in range(len(points)) if i not in nearbyPointIndices[lineSegmentPointIndices]], dtype=np.int64)
            #print(f'removed {len(points) - len(keepIndices)} points')
            points = points[keepIndices]

        ##############################################
        #    Removing votes from used points
        ##############################################
        # Now, we want to have all of the points
        # involved in this fit to "unvote" in the hough space,
        # such that we can find the next best line
        for k in range(len(directionVectors)):
            b = directionVectors[k]
            # We can compute a vector in advance that we just dot with
            # each point to find the x' and y'
            # See equation 2 in Dalitz et al. 2017
            reducedRepMultX, reducedRepMultY = reducedRepConversionMatrices(b)

            # I'm not 100% clear on why adding the system length scale is
            # necessary here, but otherwise most of our lattice represents
            # values of x' and y' that aren't relevant...
            xPrime = np.dot(reducedRepMultX, nearbyPoints.T) + systemLengthScale/2
            yPrime = np.dot(reducedRepMultY, nearbyPoints.T) + systemLengthScale/2
            #xPrime = np.dot(reducedRepMultX, points.T) + systemLengthScale/2
            #yPrime = np.dot(reducedRepMultY, points.T) + systemLengthScale/2

            # Now convert to an index, so we can find the corresponding
            # place on the hough space lattice.
            xIndex = np.floor(xPrime / stepSize).astype(np.int32)
            yIndex = np.floor(yPrime / stepSize).astype(np.int32)

            # Only take the indices that are on the lattice
            validIndices = np.where((xIndex < latticeSize) & (yIndex < latticeSize))[0]

            #houghSpace[xIndex[validIndices],yIndex[validIndices], np.repeat(k,len(validIndices)).astype(np.int32)] -= 1
            for j in validIndices:
                houghSpace[xIndex[j], yIndex[j], k] -= 1

            #print(np.unique(np.where(houghSpace - oldHough < 0)[-1], return_counts=True))

            ##############################################
            #    DEBUG
            ##############################################
            # Uncomment these to plot each line after it is detected.
            # Doesn't work with numba, so you'll have to remove the annotation.

            # if len(lineSegmentPoints) > minPointsPerLine:
            #     flattenedLinePoints = linePointArr.reshape((len(linePointArr)*2, 3))

            #     lineSet = o3d.geometry.LineSet()
            #     lineSet.points = o3d.utility.Vector3dVector(flattenedLinePoints)
            #     lineSet.lines = o3d.utility.Vector2iVector([[i,i+1] for i in range(len(flattenedLinePoints))[::2]])

            #     pcd = o3d.geometry.PointCloud()
            #     pcd.points = o3d.utility.Vector3dVector(lineSegmentPoints)

            #     pcd2 = o3d.geometry.PointCloud()
            #     pcd2.points = o3d.utility.Vector3dVector(points)
            #     pcd2.paint_uniform_color([0.4, 0.4, 0.4])

            #     o3d.visualization.draw_geometries([lineSet, pcd2, pcd])
            ##############################################
            #    DEBUG
            ##############################################


    # Remove the first entry, since that was a dummy entry
    return linePointArr[1:] + pointCloudTranslation



import numpy as np

import numba

from scipy.signal import convolve


def courseGrainField(points, values=None, defaultValue=0, latticeSpacing=None, defaultLatticeSize=100, fixedBounds=None, kernel='gaussian', kernelSize=5, subsample=None, returnSpacing=False, returnCorner=False):
    """
    Course grains a collection of values at arbitrary points,
    into a discrete field.

    If `values=None`, course-grained field is the point density.

    Parameters
    ----------
    points : numpy.ndarray[N,d]
        Spatial positions of N points in d-dimensional space.

    values : numpy.ndarray[N,[k]] or func(points)->numpy.ndarray[N,[k]] or None
        Field values at each point. Can be k-dimensional vector,
        resulting in k course-grained fields.

        Can also be a (vectorized) function that returns a value given
        a collection of points. eg. neighbor counting function. This
        functionality is provided such that if the function is computationally
        expensive, eg. neighbor counting, the points can be subdivided into
        batches and the course grained fields can be summed at the end. This
        is a way to approximate the course grained field for a huge (>1e6)
        number of points, while still remaining computationally feasible.
        See `subsample`.

        If `None`, returned field will be the point density.

    defaultValue : float or numpy.ndarray[k]
        The default value of the course-grained field;
        probably `0` for most applications.

    latticeSpacing : float or None
        The spacing of lattice points for the course-grained field.

        If `None`, will be chosen such that each axis has
        `defaultLatticeSize` points.

    defaultLatticeSize : int
        The number of lattice points for the course grained field, assuming
        no explicit value for the lattice spacing is given (see `latticeSpacing`).

    fixedBounds : numpy.ndarray[d] or None
        The bounds of the field to define the discretized
        grid over. If None, will be calculated based on the
        extrema of the provided points.

    kernel : str or numpy.ndarray[A,A]
        The kernel to course-grain the field with. 'gaussian'
        option is implemented as default, but a custom matrix
        can be provided. If using default gaussian option,
        kernel size can be set with `kernelSize`.

    kernelSize : int
        The kernel size to use if `kernel='gaussian'`.
        If a custom kernel is provided, this has no effect.

    returnSpacing : bool

    returnCorner : bool
    """
    # TODO: Make sure this works for 1D data
    dim = np.shape(points)[-1] if len(np.shape(points)) > 1 else 1

    if dim == 1:
        points = np.array(points)[:,None]
    
    if not hasattr(fixedBounds, '__iter__'):
        occupiedVolumeBounds = np.array(list(zip(np.min(points, axis=0), np.max(points, axis=0))))
    else:
        occupiedVolumeBounds = np.array(fixedBounds)
    
    # Create a lattice with the selected scale for that cube
    if latticeSpacing is not None:
        spacing = latticeSpacing
        # We also have to correct the occupied volume bounds if we were provided with
        # a fixed set of bounds. Otherwise, we will end up with an extra bin at the
        # end
        if hasattr(fixedBounds, '__iter__'):
            occupiedVolumeBounds[:,1] -= spacing
    else:
        # Choose such that each axis has 100 lattice points (-1 because we will add one later)
        spacing = (occupiedVolumeBounds[:,1] - occupiedVolumeBounds[:,0]) / (defaultLatticeSize-1)

    # In the exceptional case that the data is given as a d dimensional array
    # but the data is actually d-1 dimensional (or d-2, etc.), we will have a value
    # of spacing for that dimension as 0, which will cause a divide by zero error
    # below. If this is the case, we only need a single entry in that dimension.
    if hasattr(spacing, '__iter__'):
        spacing[spacing == 0] = 1
    else:
        spacing = spacing if spacing != 0 else 1
    
    fieldDims = (np.ceil(1 + (occupiedVolumeBounds[:,1] - occupiedVolumeBounds[:,0])/(spacing))).astype(np.int64)

    latticePositions = np.floor((points - occupiedVolumeBounds[:,0])/(spacing)).astype(np.int64)

    # Check if an array of values was passed for each point
    # Otherwise we just have a scalar field (and we'll collapse
    # the last dimension later on).
    if hasattr(values, '__iter__'):
        k = np.shape(values)[-1]
        valArr = values
    else:
        k = 1
        valArr = np.ones((np.shape(points)[0], 1))

    fieldArr = np.zeros((*fieldDims, k))
    # Instead of actually applying a gaussian kernel now, which would be
    # very inefficient since we'd need to sum a potentially very large number
    # of k*d dimensional matrices (more or less), we instead just assign each
    # lattice point, then smooth over it after with the specified kernel.
    # Where this might cause issues:
    # - If the lattice spacing is too large, you will get some weird artifacts
    #   from this process. Though in that case, you'll get a ton of artifacts from
    #   elsewhere too, so just don't use too large a lattice spacing :)
    #print(tuple(latticePositions[0]))
    for i in range(np.shape(points)[0]):
        fieldArr[tuple(latticePositions[i])] += valArr[i]

    # Now smooth over the field
    if kernel == 'gaussian':
        gaussianBlurKernel = np.zeros(np.repeat(kernelSize, np.shape(points)[-1]))
        singleAxis = np.arange(kernelSize)
        kernelGrid = np.meshgrid(*np.repeat([singleAxis], np.shape(points)[-1], axis=0))
        #kernelGrid = np.meshgrid(singleAxis, singleAxis, singleAxis)
        # No 2 prefactor in the gaussian denominator because I want the kernel to
        # decay nearly to 0 at the corners
        kernelArr = np.exp(-np.sum([(kernelGrid[i] - (kernelSize-1)/2.)**2 for i in range(np.shape(points)[-1])], axis=0) / (kernelSize))
        # Now account for however many dimensions k we have
        #kernelArr = np.repeat([kernelArr] if k > 1 else kernelArr, k, axis=0)

    # Otherwise, we expect that kernel should already be passed as a
    # proper square d-dimensional matrix
    else:
        kernelArr = kernel

    # Perform a convolution of the field with our kernel
    # 'same' keeps the same bounds on the field, but might cause
    # some weird effects near the boundaries
    # Divide out the sum of the kernel to normalize
    transConvolution = np.zeros_like(fieldArr.T)

    for i in range(k):
        # Note that convolve(x, y) == convolve(x.T, y.T).T
        # We need this so we can go over our k axis
        transConvolution[i] = convolve(fieldArr.T[i], kernelArr.T, mode='same') / np.sum(kernelArr)

    convolution = transConvolution.T

    # If k == 1, collapse the extra dimension
    if k == 1:
        convolution = convolution[..., 0]
    
    returnResult = [convolution]

    if returnSpacing:
        returnResult += [spacing]

    if returnCorner:
        returnResult += [occupiedVolumeBounds[:,0]]

    return returnResult if len(returnResult) > 1 else convolution


@numba.njit()
def pathIntegralAlongField3D(field, path, latticeSpacing=np.array([1.,1.,1.]), fieldOffset=None):
    """
    numba-optimized function to extract values along a line through
    a 3D field. Can be easily modified to compute a path integral
    (noted in comments).

    Parameters
    ----------
    field : numpy.ndarray[N,M,P]
        Field over which to compute the path integral.

    path : numpy.ndarray[L,3]
        L ordered points representing a path
        through the field.

    latticeSpacing : numpy.ndarray[3]
        The lattice spacing for the discretized field;
        can be a single value for all dimensions, or different
        values for each dimension.

    fieldOffset : numpy.ndarray[3] or None
        The position of the bottom left corner
        of the discrete lattice on which the field exists.

    Returns
    -------
    pathCut : numpy.ndarray[L]

    """
    # This was written in a more general case,
    # so some of the code will look like it isn't specific
    # to 3D, but it has been made so here.
    d = field.ndim

    # Scale the path to have no units
    scaledPath = path.astype(np.float64)

    if fieldOffset is not None:
        scaledPath -= fieldOffset

    for i in range(d):
        scaledPath[:,i] /= latticeSpacing[i]

    nearbyIndices = []
    for i in range(len(scaledPath)):
        possibleIndices = []
        for j in range(d):
            belowIndex = np.floor(scaledPath[i,j])
            aboveIndex = np.ceil(scaledPath[i,j])
           
            # So this should just be [] but there is a chance
            # that this list could contain no elements, which numba does
            # not like at all. So we have to give some indication of what
            # type this list will hold, without actually giving it any elements.
            possibilities = [0 for _ in range(0)]

            # Make sure that the indices are valid
            if belowIndex >= 0 and belowIndex < field.shape[j]:
                possibilities += [belowIndex]
            if aboveIndex >= 0 and aboveIndex < field.shape[j]:
                possibilities += [aboveIndex]
            
            possibleIndices.append(np.array(possibilities))
            #possibleIndices.append(np.array([np.floor(scaledPath[i,j]), np.ceil(scaledPath[i,j])]))
        
        # Count up how many unique sets of indices we can create from the
        # possibilities for each axis
        #totalCombinations = int(np.nanprod(np.array([np.float64(len(p)) for p in possibleIndices])))
        totalCombinations = len(possibleIndices[0])*len(possibleIndices[1])*len(possibleIndices[2]) 

        # Have to manually index to make numba happy
        # TODO: Make sure this works if the z dimension is near the edge
        # I have tested it for x and y, but not the other one yet... if you have issues
        # around this area, it might be that.
        result = np.zeros((totalCombinations, d))
        result[:,0] = np.repeat(possibleIndices[0], len(possibleIndices[1])*len(possibleIndices[2]))
        result[:,1] = list(possibleIndices[1]) * len(possibleIndices[0])*len(possibleIndices[2])
        result[:,2] = list(np.repeat(possibleIndices[2], len(possibleIndices[1]))) * len(possibleIndices[0])
        
        #print(result)
        nearbyIndices.append(result)

    fieldValuesAlongPath = np.zeros(len(scaledPath))

    for i in range(len(scaledPath)):
        localPoints = nearbyIndices[i]
        if len(localPoints) == 0:
            continue

        # Compute distances to each nearby point
        # Add some tiny amount to avoid divide by zero issues
        localDistances = np.sqrt(np.sum((scaledPath[i] - localPoints)**2, axis=-1)) + 1e-8

        interpolationContributions = localDistances / np.sum(localDistances)

        #print(localPoints)
        # Have to do some weird indexing to make numba happy, but generally this is
        # just the dot product between interpolationContributions and field[all indices]
        for j in range(len(localPoints)):
            #index = (localPoints[j][0], localPoints[j][1], localPoints[j][2])
            fieldValuesAlongPath[i] += interpolationContributions[j] * field[int(localPoints[j][0]), int(localPoints[j][1]), int(localPoints[j][2])]

    return fieldValuesAlongPath
    # To get the actual path integral value, you can remove this
    # return statement

    # We need to weigh our numerical integration by the step size
    # We just to do a centered step scheme. ie. the interpolated value computed
    # above for point i "counts" for half of the path approaching point i, and
    # half of the path leaving point i. Thus, the first and last point are weighed
    # only half, since they don't have an incoming or outgoing path each.
    # We also have to scale back to the original lattice spacing.
    unscaledPath = scaledPath[1:] - scaledPath[:-1]
    for i in range(d):
        unscaledPath[:,i] *= latticeSpacing[i]

    pathSpacings = np.sqrt(np.sum((unscaledPath)**2, axis=-1))
    symSpacing = (pathSpacings[:-1] + pathSpacings[1:])/2
    symSpacing = np.concatenate((np.array([pathSpacings[0]/2]), symSpacing, np.array([pathSpacings[-1]/2])))

    pathIntegral = np.sum(symSpacing*fieldValuesAlongPath)

    return pathIntegral



import numpy as np
import numba

from .utils import reducedRepConversionMatrices, distancePointToLine, unravel_3d_index

@numba.njit(cache=True)
def hough3D(points, directionVectors, latticeSize=128, neighborDistance=0.01, minPointsPerLine=5):
    """
    Perform line detection on a 3D point cloud.

    This version of the function is follows the algorithm described in
    the original paper [1] (except for step 5):
        1. Have all points vote in the hough space.
        2. Take the highest voted cell as an initial guess for a new line.
        3. Find all points less than a certain distance to the initial line.
        4. Perform linear regression on the nearby points to refine the line.
        5. Update which points are nearby to the line based on the updated line (NEW).
        6. Have all of the points that are part of this line reverse their vote.
        7. Save the information about the detected line.
        8. Repeat steps 2-7 until there are not enough points left to form a line.

    I found that adding step 5 gave slightly more reliable detections, so
    I have included it. Note that you could take this further, and perform
    several steps of fitting and updating, but here we just do one.
    
    Parameters
    ----------
    points : numpy.ndarray[N,3]
        Points representing the point cloud.

    directionVectors : numpy.ndarray[M,3]
        Direction vectors represented the discretized directions for
        the lines. Normally generated with `genIcosahedron()`.

    latticeSize : int
        The number of discrete boxes to use for the intercept variables in
        hough space.

    neighborDistance : float
        The distance a point can be from the initial guess of a
        line for it to be included in the linear regression
        used to refine the line parameters. Given as fraction of
        the total system size (diagonal length of the bounding box).

    minPointsPerLine : int
        The minimum number of points that must be part of a line
        detection for that line to be included in the final result.

    Returns
    -------
    linePointArr : numpy.ndarray[K,2,3]
        The position vectors for the start and end of each
        line.

    """
    ##############################################
    #              Preparation steps
    ##############################################
    # Shift the mean of the point cloud to the origin
    pointCloudTranslation = np.zeros(3)
    for i in range(3):
        pointCloudTranslation[i] = np.mean(points[:,i])
        
    points = np.copy(points) - pointCloudTranslation
    
    # Compute the discrete direction vectors for our lines
    #directionVectors = genIcosahedron(directionGranularity)
    # TODO: Make this generation function numba compatable
    
    # Find the extents of the point cloud
    bounds = np.zeros((2, 3))
    for i in range(3):
        bounds[:,i] = [np.min(points[:,i]), np.max(points[:,i])]
    #bounds = np.array([ for i in range(3)]).T

    systemLengthScale = np.sqrt(np.sum((bounds[1] - bounds[0])**2))

    stepSize = systemLengthScale / latticeSize

    # Our hough space will be a three dimensional array, with the first two
    # dimensions representing the reduced coordinates x' and y' for the
    # intersection point of a line with a plane that passes through [0,0,0],
    # and the last dimension for the direction vector of the line. Altogether,
    # a point i,j,k in the hough space fully describes a line in the space
    # of the point cloud.
    houghSpaceSize = (latticeSize, latticeSize, len(directionVectors))

    ##############################################
    #    Initial transformation to hough space
    ##############################################
    # The actual hough space where points will be voting
    houghSpace = np.zeros(houghSpaceSize)

    # Now do the voting
    for k in range(len(directionVectors)):
        b = directionVectors[k]
        # We can compute a vector in advance that we just dot with
        # each point to find the x' and y'
        # See equation 2 in Dalitz et al. 2017
        reducedRepMultX, reducedRepMultY = reducedRepConversionMatrices(b)

        # I'm not 100% clear on why adding the system length scale is
        # necessary here, but otherwise most of our lattice represents
        # values of x' and y' that aren't relevant...
        xPrime = np.dot(reducedRepMultX, points.T) + systemLengthScale/2
        yPrime = np.dot(reducedRepMultY, points.T) + systemLengthScale/2

        # Now convert to an index, so we can find the corresponding
        # place on the hough space lattice.
        xIndex = np.floor(xPrime / stepSize).astype(np.int32)
        yIndex = np.floor(yPrime / stepSize).astype(np.int32)

        # Get rid of indices that are not on our lattice
        validIndices = np.where((xIndex < latticeSize) & (yIndex < latticeSize))[0]

        #print(validIndices)
        for j in validIndices:
            houghSpace[xIndex[j], yIndex[j], k] += 1

    ##############################################
    #    Iterative selection of lines
    ##############################################
    # Now that we have calulated our hough space, we need to identify
    # maxima in it to find the best lines

    # We do this by starting with the most voted point in hough space
    # and then "unvote" all of the points that are associated with that
    # line, and then repeat the process iteratively
    
    # Numba doesn't play well with empty arrays because
    # it can't infer the type of the data, so we just start
    # with an empty line (all zeros) and we will remove
    # it at the end (see return statement).
    # First dimension indexes the line, second the start or end point
    # of the line, and the final dimension is for the space we are in.
    linePointArr = np.zeros((1, 2, 3))

    while True:
        # The highest voted space
        maxIndex = np.argmax(houghSpace)
        # Now we have to unravel the index, since it is given as a 1D index
        # The function numpy.unravel_index() does this, but it is not supported
        # by numba
        maxIndex = unravel_3d_index(maxIndex, houghSpaceSize)

        # Find the information about the line that this index
        # represents in hough space
        xIndex, yIndex, bIndex = maxIndex

        # DEBUG
        #plt.imshow(houghSpace[:,:,bIndex])
        #plt.colorbar()
        #plt.show()

        ##############################################
        #    Convert from hough space to real space
        ##############################################
        # Converted from index to reduced coordinates
        xPrime = xIndex*stepSize - systemLengthScale/2
        yPrime = yIndex*stepSize - systemLengthScale/2

        # Calculate the anchor point on the line, which,
        # together with the direction vector, fully identifies
        # the line.
        b = directionVectors[bIndex]
        reducedRepMultX, reducedRepMultY = reducedRepConversionMatrices(b)
        # See equation 3 from Dalitz et al. 2017
        anchorPoint = xPrime*reducedRepMultX + yPrime*reducedRepMultY

        # Now, we need to refine the line using linear regression.
        # First, that means identifying which points are included in
        # this line.

        # Compute the line intersection of the given line and
        # a perpendicular one that passes through every point, and then
        # find the distance from intersection to the respective point.
        distancesToPoints = np.array([distancePointToLine(p, anchorPoint, b) for p in points])
        nearbyPointIndices = np.where(distancesToPoints < systemLengthScale*neighborDistance)[0]

        # Make sure we have enough points to form a line
        nearbyPoints = points[nearbyPointIndices]

        if len(nearbyPoints) < minPointsPerLine:
            break

        ##############################################
        #    Linear regression on nearby points
        ##############################################
        # Now we perform linear regression on those points to fit a line
        # and increase the accuracy of our detected line

        # We will use SVD to find the first principle component
        # of the local points, which is equivalent to a linear
        # fit.
        # To do this, we need our data to be zero-centered
        # (and we need the center of mass point anyway as our
        # anchor point)
        anchorPoint = np.zeros(3)
        for i in range(3):
            anchorPoint[i] = np.mean(nearbyPoints[:,i])

        # SVD on the mean-centered data.
        # The original paper did PCA instead of SVD, as they said it
        # was much faster, but I think SVD is fine here...
        uu, dd, vv = np.linalg.svd(nearbyPoints - anchorPoint)
        # Direction vector is the first principle component
        directionVector = vv[0]

        # Update which points are close to the line based on
        # the better fit
        # NOTE: this is different from the original paper, as they do not
        # perform this step. I find it helps, so I've included it.
        distancesToPoints = np.array([distancePointToLine(p, anchorPoint, directionVector) for p in points])
        nearbyPointIndices = np.where(distancesToPoints < systemLengthScale*neighborDistance)[0]
        nearbyPoints = points[nearbyPointIndices]

        if len(nearbyPoints) < minPointsPerLine:
            break

        for i in range(3):
            anchorPoint[i] = np.mean(nearbyPoints[:,i])
           
        # Compute the length of the line
        # The best way to do this is to compute the parametric representation
        # of every point in nearbyPoints. Then we can take the maximum and
        # minimum parametric values, to find how far the line extends in each direction
        tArr = np.dot(directionVector, (nearbyPoints - anchorPoint).T)
        lineStart = anchorPoint + np.max(tArr)*directionVector
        lineEnd = anchorPoint + np.min(tArr)*directionVector

        linePoints = np.zeros((2, 3))
        linePoints[0] = lineStart
        linePoints[1] = lineEnd

        # Save the information about this line
        linePointArr = np.concatenate((linePointArr, linePoints[None,:,:]))
        
        ##############################################
        #    Removing votes from used points
        ##############################################
        # Now, we want to have all of the points
        # involved in this fit to "unvote" in the hough space,
        # such that we can find the next best line
        for k in range(len(directionVectors)):
            b = directionVectors[k]
            # We can compute a vector in advance that we just dot with
            # each point to find the x' and y'
            # See equation 2 in Dalitz et al. 2017
            reducedRepMultX, reducedRepMultY = reducedRepConversionMatrices(b)
    
            # I'm not 100% clear on why adding the system length scale is
            # necessary here, but otherwise most of our lattice represents
            # values of x' and y' that aren't relevant...
            xPrime = np.dot(reducedRepMultX, nearbyPoints.T) + systemLengthScale/2
            yPrime = np.dot(reducedRepMultY, nearbyPoints.T) + systemLengthScale/2
            #xPrime = np.dot(reducedRepMultX, points.T) + systemLengthScale/2
            #yPrime = np.dot(reducedRepMultY, points.T) + systemLengthScale/2
    
            # Now convert to an index, so we can find the corresponding
            # place on the hough space lattice.
            xIndex = np.floor(xPrime / stepSize).astype(np.uint8)
            yIndex = np.floor(yPrime / stepSize).astype(np.uint8)
    
            # Get rid of indices that are not on our lattice
            validIndices = np.where((xIndex < latticeSize) & (yIndex < latticeSize))[0]
            
            #print(validIndices)
            for j in validIndices:
                houghSpace[xIndex[j], yIndex[j], k] -= 1

        # Remove the used points from the list
        keepIndices = np.array([i for i in range(len(points)) if i not in nearbyPointIndices], dtype=np.int64)
        points = points[keepIndices]
        
        ##############################################
        #    DEBUG
        ##############################################
        # Uncomment these to plot each line after it is detected.
        # Doesn't work with numba, so you'll have to remove the annotation.
        
        # import open3d as o3d
        
        # flattenedLinePoints = linePointArr.reshape((len(linePointArr)*2, 3))
        
        # lineSet = o3d.geometry.LineSet()
        # lineSet.points = o3d.utility.Vector3dVector(flattenedLinePoints)
        # lineSet.lines = o3d.utility.Vector2iVector([[i,i+1] for i in range(len(flattenedLinePoints))[::2]])
        
        # pcd = o3d.geometry.PointCloud()
        # pcd.points = o3d.utility.Vector3dVector(nearbyPoints)
        
        # o3d.visualization.draw_geometries([lineSet, pcd])
        ##############################################
        #    DEBUG
        ##############################################

    # Remove the first entry, since that was a dummy entry
    # And we have to translate the point could back to its original position
    return linePointArr[1:] + pointCloudTranslation



import numpy as np

from itertools import product

from scipy.spatial import KDTree

import numba

def genIcosahedron(nSubdivisions=4):
    """
    Generate direction vectors to the vertices of
    a icosahedron subdivided a given amount of times,
    to create a discrete set direction vectors equally
    sampling space.

    Due to not being able to figure out the analytic
    expression for the distance between connected components
    on the shape, only works up to 5 subdivisions. This is
    a hard problem because we can't quite just take the nearest
    neighbor distance because a icosahedron has some vertices
    with 5 neighbors (spaced a distance `a` away) and some
    vertices with 6 neighbors (spaced a distance `b` away, `a`!=`b`).

    Subdivisions | Directions
        0              6
        1              21
        2              81
        3              321
        4              1281
        5              5121

    Parameters
    ----------
    nSubdivisions : int
        The number of times to subdivide the shape, creating a finer
        grid of directions. See table above for number of directions.

    Returns
    -------
    directionVectors : numpy.ndarray[N,3]

    """
    # First, generate the basic icosahedron
    vertices = []
    r = 1/2 * (1 + np.sqrt(5))

    # See equation 4 of Dalitz et al. (2017)
    # This generates all combinations of:
    # (+-1, +-r, 0)
    # (0, +-1, +-r)
    # (+-r, 0, +-1)
    vertexPossibilities = np.array([[1, -1], [0], [r, -r]], dtype=list)

    for i in range(len(vertexPossibilities)):
        rolled = np.roll(vertexPossibilities, i)
        xPossibilities = rolled[0]
        yPossibilities = rolled[1]
        zPossibilities = rolled[2]

        vertices += list(product(xPossibilities, yPossibilities, zPossibilities))

    # Now we have an array of 12 vertices for our basic shape
    vertices = np.array(vertices)

    # Next, we add a vertex between adjacent vertices the
    # requested number of times

    # These were calculated manually, by finding the distance required to make
    # sure every vertex has 5 or 6 neighbors... I'm sure there must be an
    # analytic expression here, but whatever, we probably would only ever
    # use up to 4 or 5 (at the absolute max) so it is what it is...
    # but still TODO
    neighborDistances = [1.0515, 0.6180, 0.3249, 0.1647, 0.0826]
    if nSubdivisions > len(neighborDistances):
        raise Exception(f'Too many subdivisions requested! Only supported up to {len(neighborDistances)}...')

    for i in range(nSubdivisions):

        # Normalize the vertices so they sit on the surface of a sphere
        norms = np.sqrt(np.sum(vertices**2, axis=-1))
        vertices = np.array([vertices[j]/norms[j] for j in range(len(vertices))])
        # Remove duplicates
        vertices = np.unique(vertices, axis=0)

        # We will determine which points are adjacent using a
        # KD Tree; points that are adjecent should be a distance
        # of 2/n away from each other, where n is the current iteration
        # beginning with n=1.
        # Every vertex should have either 5 or 6 neighbors

        # This was used for manually finding the neighbor distances
        #uniq = np.array(np.unique(np.array([np.sqrt(np.sum((vertices - vertices[j])**2, axis=-1)) for j in range(len(vertices))]).flatten(), return_counts=True)).T
        #for u,c in uniq:
        #    print(u, int(c))

        kdTree = KDTree(vertices)
        # Add some small epsilon just in case
        neighbors = kdTree.query_ball_tree(kdTree, neighborDistances[i]*1.15+1e-6)

        # The first point will always be the same point itself, so
        # we need to throw that away
        neighbors = [n[1:] for n in neighbors]
        #print([len(n) for n in neighbors])

        # Now add points between each point and its neighbors
        newVertices = np.array([(np.array(vertices[j]) + np.array(vertices[k]))/2 for j in range(len(neighbors)) for k in neighbors[j]])

        # Add the new vertices to the running array
        vertices = np.concatenate([vertices, newVertices])

    # Final normalization of the vertices
    norms = np.sqrt(np.sum(vertices**2, axis=-1))
    vertices = np.array([vertices[j]/norms[j] for j in range(len(vertices))])

    vertices = np.unique(vertices, axis=0)

    # Now we have to remove mirror images of directions
    # eg. [0, 0, 1] and [0, 0, -1] represent antiparallel
    # slopes, but for our fitting, this difference doesn't matter
    # We do this by marking points as duplicate by reflecting
    # points across each axis, one at a time
    keepVertex = np.ones(len(vertices), dtype=bool)
    for i in range(len(vertices)):
        # We have to first set this to false such that we don't
        # compute the dot product of the vector with itself
        keepVertex[i] = False
        # Now we compute the dot products and make sure none of them
        # are -1 (ie. there exists an antiparallel vector in the list already)
        # Need a small epsilon to account for floating point errors
        keepVertex[i] = np.min(np.dot(vertices[i], vertices[keepVertex].T)) > -1 + 1e-8

    vertices = vertices[keepVertex]

    return vertices


@numba.njit()
def reducedRepConversionMatrices(b):
    """
    Calculate the matrix to convert a line representation
    to it's reduced represetation [1] as:
        ```
        X, Y = reducedRepConversionMatrices(b)
        x' = np.dot(X, p)
        y' = np.dot(Y, p)
        ```

    Parameters
    ----------
    b : numpy.ndarray[3]
        The vector pointing in the direction of the line for which
        we want to calculate the conversion matrices

    Returns
    -------
    X : numpy.ndarray[3]

    Y : numpy.ndarray[3]

    References
    ----------
    [1] K.S. Roberts, A new representation for a line, in Computer Vision
    and Pattern Recognition CVPR’88, 1988, pp. 635–640.

    """
    reducedRepMultX = np.array([1 - b[0]**2/(1 + b[2]),
                               -b[0]*b[1]/(1 + b[2]),
                               -b[0]])

    reducedRepMultY = np.array([-b[0]*b[1]/(1 + b[2]),
                                1 - b[1]**2/(1 + b[2]),
                               -b[1]])

    return reducedRepMultX, reducedRepMultY


@numba.njit()
def rotationMatrix(theta, phi, psi):
    """
    Generate the rotation matrix corresponding to rotating
    a point in 3D space.
    """
    return np.array([[np.cos(theta)*np.cos(psi), np.cos(phi)*np.sin(psi) + np.sin(phi)*np.sin(theta)*np.cos(psi), np.sin(phi)*np.sin(psi) - np.cos(psi)*np.cos(phi)*np.sin(theta)],
                     [-np.cos(theta)*np.sin(psi), np.cos(phi)*np.cos(psi) + np.sin(phi)*np.sin(theta)*np.sin(psi), np.sin(phi)*np.cos(psi) - np.cos(psi)*np.sin(phi)*np.sin(theta)],
                     [np.sin(theta), -np.sin(phi)*np.cos(theta), np.cos(phi)*np.cos(theta)]])


@numba.njit()
def lineIntersection(a1, b1, a2, b2):
    """
    Compute the intersection of two lines:
        a1 + b1*t and a2 + b2*t
    """
    # Compute the cross product
    crossProd = np.cross(b1, b2)
    if crossProd == 0:
        return np.nan
    # Dot the cross product unit vector into the vector from intercept to intercept
    distance = np.dot(a2 - a1, crossProd) / np.sqrt(np.sum(crossProd**2))
    return distance


@numba.njit()
def distancePointToLine(p, a, b):
    """
    Compute intersection of a point (p) with a line (a + bt).

    Parameters
    ----------
    p : numpy.ndarray[3]
        Point to compute distance to the line

    a : numpy.ndarray[3]
        Point located on the line.

    b : numpy.ndarray[3]
        Direction vector of the line.

    Returns
    -------
    d : float
        Distance from the point to the closest point
        on the line.
    """
    t = np.dot(b, p - a)
    closestPointOnLine = a + t*b
    return np.sqrt(np.sum((p - closestPointOnLine)**2))


@numba.njit()
def unravel_3d_index(index, shape):
    """
    Turn a 1D index into a 3D index.

    Equivalent to numpy.unravel_index specifically for
    3D; numpy's function isn't supported by numba, so I
    had to write my own.

    Parameters
    ----------
    index : int
        1D index for an array; for example, generated by
        `np.argmax(arr)`.

    shape : list[3]
        Size of each dimension of the array you want to index.

    Returns
    -------
    rIndex : list[3]
    """
    properIndex = np.zeros(3, dtype=np.int64)

    properIndex[0] = np.floor(index / (shape[1]*shape[2]));
    properIndex[1] = np.floor((index - properIndex[0] * shape[1]*shape[2]) / shape[2]);
    properIndex[2] = index - shape[2] * (properIndex[1] + shape[1] * properIndex[0]);

    return properIndex

def animateGeometries(geometries, outputFile=None, rotationSpeed=5, maxFrames=None, fps=30, pointSize=7, lineWidth=5, zoom=0.8, backgroundColor=[1,1,1], crop=False, cropPadding=10, downsample=1, outputFormat='gif'):
    """
    Create an animation of a set of geometries rotating.

    Uses Open3D as the graphical framework.

    Parameters
    ----------
    geometries : list(open3d.geometry.Geometry)
        The list of geometries to draw, eg. point clouds, line sets, etc.

    outputFile : str or None
        The name of the animation file that will be saved. If it includes
        an (acceptable) extension, the output type will be automatically
        chosen; otherwise, the output format will be set by the
        kwarg `outputFormat`.

        If `None`, no file will be saved, the animation will just be
        played.

    rotationSpeed : int
        The speed of the rotation during the animation.

    maxFrames : int or None
        The total number of frames to save; if `None`, no limit will
        be applied.

        Note that I can't figure out how to exit a window automatically,
        so the window itself will continue existing until the user manually
        presses 'q', but no more than `maxFrames` will actually be saved.

    fps : int
        The frames per second of the output animation.

    pointSize : float
        The size of points for point clouds.

    lineWidth : float
        The width of lines for line sets (WIP).

    zoom : float
        The zoom setting for the view.

    backgroundColor : [float, float, float]
        The background color for the animation.

    crop : bool
        Whether to crop the animation to a minimal window size (plus
        padding) that only includes pixels that actually change
        during the animation.

    cropPadding : int
        The amount of padding to add around the animation if
        `crop=True`.

    outputFormat : 'gif' or 'mp4'
        The output format for the animation. Overridden if an extension is
        directly included in the file name (`outputFile`).
        
    """
    # Import these here since they are needed in the rest of this file
    # and this is sortof a debug method
    import cv2
    from PIL import Image
    import open3d as o3d

    ALLOWED_FORMATS = ['gif', 'mp4']

    if outputFile is not None:
        # Try to infer the output format from the file name
        potentialExtension = outputFile.split('.')[-1].lower()
        if potentialExtension in ALLOWED_FORMATS:
            outputFormat = potentialExtension
            fullOutputPath = outputFile
        else:
            # Otherwise, we have to append whatever the provided extension is
            outputFormat = outputFormat.lower()

            # But first make sure the provided format is allowed
            if outputFormat not in ALLOWED_FORMATS:
                raise Exception(f'Invalid output format specified: {outputFormat}; available options are {ALLOWED_FORMATS}.')

            fullOutputPath = f'{outputFile}.{outputFormat}'

    images = []

    def update_view(vis):
        opt = vis.get_render_option()
        opt.background_color = np.asarray(backgroundColor)

        vis.get_render_option().point_size = pointSize
        vis.get_render_option().line_width = lineWidth


        ctr = vis.get_view_control()
        ctr.rotate(rotationSpeed, 0)
        ctr.set_zoom(zoom)

        if outputFile is not None and (maxFrames is None or len(images) < maxFrames):
            image = (np.array(vis.capture_screen_float_buffer(False))*255).astype(np.uint8)
            images.append((image).astype(np.uint8))

        return False

    o3d.visualization.draw_geometries_with_animation_callback(geometries,
                                                              update_view)

    if outputFile is None:
        return


    if crop:
        # If we want to crop, we should see if there are any pixels that only
        # ever take on the value of the background color
        usedPixels = np.sum(np.array(images) - np.array(backgroundColor)*255, axis=0)
        usedPixels = np.sum(usedPixels**2, axis=-1)

        usedIndices = np.array(np.where(usedPixels > 1e-8))
        cropBounds = [max(np.min(usedIndices[0])-cropPadding, 0),
                      min(np.max(usedIndices[0])+cropPadding, images[0].shape[0]),
                      max(np.min(usedIndices[1])-cropPadding, 0),
                      min(np.max(usedIndices[1])+cropPadding, images[0].shape[1])]

    else:
        cropBounds = [0, images[0].shape[1], 0, images[0].shape[0]]

    for i in range(len(images)):
        images[i] = images[i][cropBounds[0]:cropBounds[1],cropBounds[2]:cropBounds[3]][::downsample,::downsample]

    if outputFormat == 'mp4':
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        video = cv2.VideoWriter(fullOutputPath, fourcc, fps, (images[0].shape[1], images[0].shape[0]))

        for image in images:
            video.write(cv2.cvtColor(image, cv2.COLOR_RGB2BGR))

        video.release()

    if outputFormat == 'gif':
        for i in range(len(images)):
            images[i] = Image.fromarray(images[i])

        images[0].save(fullOutputPath, save_all=True, append_images=images[1:], loop=0, duration=int(1/fps*1e3))




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name='hough3d',  
     version='1.0.0',
     author="Jack Featherstone",
     author_email="jack.featherstone@oist.jp",
     license='MIT',
     url='https://github.com/jfeatherstone/hough3d',
     description="Library for detecting lines in a 3D point cloud via Hough transform.",
     long_description=long_description,
     long_description_content_type="text/markdown",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3.11",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
 )
